# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utilities."""

from flask import session
from flask_principal import Identity
from invenio_accounts.models import Role
from invenio_accounts.proxies import current_db_change_history

from .generators import CommunityRoleNeed
from .proxies import current_communities, current_identities_cache

IDENTITY_KEY = "user-communities:"


def load_community_needs(identity):
    """Add community-related needs to the freshly loaded identity.

    Note that this function is intended to be called as handler for the
    identity-loaded signal, where we don't have control over the handler
    execution order.
    Thus, the given identity may not be fully initialized and still missing
    some needs (e.g. 'authenticated_user'), so we cannot rely on high-level
    service functions because permission checks may fail.
    """
    if identity.id is None:
        # no user is logged in
        return

    # Cache keys
    #
    # The cache of communities must be invalidated on:
    # 1) on creation of a community (likely this one is modelled via membership
    #    in the future).
    # 2) add/remove/change of membership
    #
    # We construct the cache key for each membership entity (e.g. user,
    # role, system role). This way, once a membership is added/removed/updated
    # we can cache the list of associated communities for this entity.
    # Once a user logs in, we get the cache for each of the membership
    # entities and combine it into a single list.

    # Currently, only users are supported (no roles or system roles)
    cache_key = identity_cache_key(identity)
    community_roles = current_identities_cache.get(cache_key)
    if community_roles is None:
        # aka Member.get_memberships(identity)
        roles_ids = session.get("unmanaged_roles_ids", [])

        member_cls = current_communities.service.members.config.record_cls
        managed_community_roles = member_cls.get_memberships(identity)
        unmanaged_community_roles = member_cls.get_memberships_from_group_ids(
            identity, roles_ids
        )
        community_roles = managed_community_roles + unmanaged_community_roles

        for community_id, role in community_roles:
            current_identities_cache.append(community_id, identity.id)
            # Add community needs to identity
            identity.provides.add(CommunityRoleNeed(community_id, role))

        current_identities_cache.set(
            cache_key,
            community_roles,
        )
    else:
        # Add community needs to identity
        for community_id, role in community_roles:
            identity.provides.add(CommunityRoleNeed(community_id, role))


def on_user_membership_change(identity=None):
    """Handler called when a user membership is changed."""
    if identity is not None:
        current_identities_cache.delete(identity_cache_key(identity))


def on_group_membership_change(community_id):
    """Handler called when a group membership is changed."""
    community_identities_cache = current_identities_cache.get(community_id)
    if community_identities_cache:
        for identity in community_identities_cache:
            on_user_membership_change(Identity(identity))


def identity_cache_key(identity):
    """Make the cache key for storing the communities for a user."""
    return f"{IDENTITY_KEY}{identity.id}"


def on_datastore_post_commit(sender, session):
    """Clears the cache for the user identity."""
    sid = id(session)
    if current_db_change_history.sessions.get(sid):
        for user_id in current_db_change_history.sessions[sid].updated_users:
            on_user_membership_change(Identity(user_id))

        for user_id in current_db_change_history.sessions[sid].deleted_users:
            on_user_membership_change(Identity(user_id))

        for role_id in current_db_change_history.sessions[sid].deleted_roles:
            role = Role.query.filter_by(id=role_id).one_or_none()
            users = role.users.all()
            for user in users:
                on_user_membership_change(Identity(user.id))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Subcommunities module service."""

from .config import SubCommunityServiceConfig
from .service import SubCommunityService

__all__ = ("SubCommunityService", "SubCommunityServiceConfig")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Test subcommunities workflows."""

import pytest
from invenio_access.permissions import system_identity
from invenio_records_resources.services.errors import PermissionDeniedError
from invenio_requests.proxies import current_requests_service

from invenio_communities.communities.records.api import Community
from invenio_communities.proxies import current_communities


@pytest.fixture(scope="module")
def subcommunity_service():
    """Return the subcommunity service."""
    return current_communities.subcommunity_service


@pytest.fixture(scope="module")
def curator(users):
    """Return the curator user."""
    return users["curator"]


@pytest.fixture(scope="module")
def reader(users):
    """Return the reader user."""
    return users["reader"]


@pytest.fixture(scope="module")
def full_community(full_community):
    """Return a full community."""
    del full_community["metadata"]["type"]
    return full_community


@pytest.fixture(scope="module")
def create_community(location, community_service):
    """Return a factory to create a community."""

    def _create_community(owner, metadata, allow_children=False):
        """Create a community."""
        c = community_service.create(owner.identity, metadata)
        if allow_children:
            c._record.children.allow = True
        c._record.commit()
        Community.index.refresh()
        owner.refresh()
        return c

    return _create_community


@pytest.fixture()
def parent_community(db, create_community, owner, minimal_community):
    """Create and return a parent community."""
    return create_community(
        owner,
        {
            **minimal_community,
            "metadata": {"title": "Parent Community"},
            "slug": "parent-community",
        },
        allow_children=True,
    )


@pytest.fixture()
def child_community(db, create_community, curator, minimal_community):
    """Create and return a child community."""
    return create_community(
        curator,
        {
            **minimal_community,
            "metadata": {"title": "Child Community"},
            "slug": "child-community",
        },
    )


def test_service_join_already_existing(
    app,
    db,
    curator,
    owner,
    parent_community,
    child_community,
    subcommunity_service,
):
    """Test joining a community with an existing subcommunity."""
    request_creator = curator
    request_receiver = owner

    payload = {"community_id": str(child_community.id)}
    result = subcommunity_service.join(
        request_creator.identity,
        str(parent_community.id),
        payload,
    )
    request_obj = result._record
    assert request_obj.status == "submitted"

    # Test accept
    current_requests_service.execute_action(
        request_receiver.identity, id_=request_obj.id, action="accept"
    )
    req_read = current_requests_service.read(request_receiver.identity, request_obj.id)
    assert req_read._request.status == "accepted"
    assert str(request_obj.topic.resolve().parent.id) == parent_community.id


def test_service_join_with_new_community(
    app,
    db,
    curator,
    owner,
    parent_community,
):
    """Test joining a community with a new subcommunity."""
    request_creator = curator
    request_receiver = owner
    payload = {
        "community": {
            "title": "Test community",
            "slug": "test-community",
        }
    }
    result = current_communities.subcommunity_service.join(
        request_creator.identity,
        str(parent_community.id),
        payload,
    )
    request_obj = result._record
    assert request_obj.status == "submitted"

    # Test accept
    current_requests_service.execute_action(
        request_receiver.identity, id_=request_obj.id, action="accept"
    )
    req_read = current_requests_service.read(system_identity, request_obj.id)
    assert req_read._request.status == "accepted"
    assert str(request_obj.topic.resolve().parent.id) == parent_community.id


def test_service_join_permissions(
    app,
    db,
    owner,
    curator,
    reader,
    parent_community,
    child_community,
):
    """Test joining with a subcommunity not owned by the user."""
    payload = {"community_id": str(child_community.id)}
    with pytest.raises(PermissionDeniedError):
        current_communities.subcommunity_service.join(
            reader.identity,
            str(parent_community.id),
            payload,
        )


# Resource tests
def test_subcommunity_simple_flow(app, curator, owner, parent_community):
    """Test the basic workflow for a subcommunity request."""
    client = curator.login(app.test_client())
    payload = {
        "community": {
            "title": "Test community",
            "slug": "test-community",
        }
    }
    # Create a request
    res = client.post(
        f"/communities/{str(parent_community.id)}/actions/join-request",
        json=payload,
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 201
    request_id = res.json["id"]
    new_community_id = res.json["topic"]["community"]
    assert res.json == {
        "id": request_id,
        "number": res.json["number"],
        "created_by": {"community": new_community_id},
        "topic": {"community": new_community_id},
        "receiver": {"community": str(parent_community.id)},
        "type": "subcommunity",
        "status": "submitted",
        "title": 'Inclusion of subcommunity "Test community"',
        "created": res.json["created"],
        "updated": res.json["updated"],
        "expires_at": None,
        "revision_id": 2,
        "is_closed": False,
        "is_expired": False,
        "is_open": True,
        "links": {
            "actions": {
                # The requester can only cancel the request
                "cancel": f"https://127.0.0.1:5000/api/requests/{request_id}/actions/cancel",
            },
            "comments": f"https://127.0.0.1:5000/api/requests/{request_id}/comments",
            "self": f"https://127.0.0.1:5000/api/requests/{request_id}",
            "self_html": f"https://127.0.0.1:5000/requests/{request_id}",
            "timeline": f"https://127.0.0.1:5000/api/requests/{request_id}/timeline",
        },
    }

    # Check the new community
    res = client.get(
        f"/communities/{new_community_id}",
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 200
    assert res.json["id"] == new_community_id
    assert res.json["metadata"]["title"] == "Test community"
    assert res.json["slug"] == "test-community"

    # Accept the request as the parent community owner
    client = owner.login(app.test_client())
    res = client.get(
        f"/requests/{request_id}",
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 200
    assert res.json["links"]["actions"] == {
        # The parent community owner can accept or decline the request
        "accept": f"https://127.0.0.1:5000/api/requests/{request_id}/actions/accept",
        "decline": f"https://127.0.0.1:5000/api/requests/{request_id}/actions/decline",
    }

    res = client.post(
        f"/requests/{request_id}/actions/accept",
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 200
    assert res.json["status"] == "accepted"
    assert res.json == {
        "id": request_id,
        "number": res.json["number"],
        "created_by": {"community": new_community_id},
        "topic": {"community": new_community_id},
        "receiver": {"community": str(parent_community.id)},
        "created": res.json["created"],
        "updated": res.json["updated"],
        "expires_at": None,
        "type": "subcommunity",
        "status": "accepted",
        "title": 'Inclusion of subcommunity "Test community"',
        "revision_id": 3,
        "is_closed": True,
        "is_expired": False,
        "is_open": False,
        "links": {
            "actions": {},
            "comments": f"https://127.0.0.1:5000/api/requests/{request_id}/comments",
            "self": f"https://127.0.0.1:5000/api/requests/{request_id}",
            "self_html": f"https://127.0.0.1:5000/requests/{request_id}",
            "timeline": f"https://127.0.0.1:5000/api/requests/{request_id}/timeline",
        },
    }

    # Check that the child community is a subcommunity of the parent community
    res = client.get(
        f"/communities/{new_community_id}",
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 200
    assert res.json["parent"]["id"] == str(parent_community.id)

    Community.index.refresh()

    res = client.get(
        f"/communities/{parent_community.id}/subcommunities",
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1


def test_subcommunity_existing_child_flow(
    app, curator, owner, parent_community, child_community
):
    """Test the subcommunity request workflow for an existing community."""
    # Create a request
    client = curator.login(app.test_client())
    res = client.post(
        f"/communities/{str(parent_community.id)}/actions/join-request",
        json={"community_id": str(child_community.id)},
        headers={"content-type": "application/json"},
    )

    assert res.status_code == 201
    request_id = res.json["id"]
    assert res.json == {
        "id": request_id,
        "number": res.json["number"],
        "created_by": {"community": str(child_community.id)},
        "topic": {"community": str(child_community.id)},
        "receiver": {"community": str(parent_community.id)},
        "type": "subcommunity",
        "status": "submitted",
        "title": 'Inclusion of subcommunity "Child Community"',
        "created": res.json["created"],
        "updated": res.json["updated"],
        "expires_at": None,
        "revision_id": 2,
        "is_closed": False,
        "is_expired": False,
        "is_open": True,
        "links": {
            "actions": {
                # The requester can only cancel the request
                "cancel": f"https://127.0.0.1:5000/api/requests/{request_id}/actions/cancel",
            },
            "comments": f"https://127.0.0.1:5000/api/requests/{request_id}/comments",
            "self": f"https://127.0.0.1:5000/api/requests/{request_id}",
            "self_html": f"https://127.0.0.1:5000/requests/{request_id}",
            "timeline": f"https://127.0.0.1:5000/api/requests/{request_id}/timeline",
        },
    }

    # Accept the request as the parent community owner
    client = owner.login(app.test_client())
    res = client.get(
        f"/requests/{request_id}",
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 200
    assert res.json["links"]["actions"] == {
        # The parent community owner can accept or decline the request
        "accept": f"https://127.0.0.1:5000/api/requests/{request_id}/actions/accept",
        "decline": f"https://127.0.0.1:5000/api/requests/{request_id}/actions/decline",
    }

    res = client.post(
        f"/requests/{request_id}/actions/accept",
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 200
    assert res.json["status"] == "accepted"
    assert res.json == {
        "id": request_id,
        "number": res.json["number"],
        "created_by": {"community": str(child_community.id)},
        "topic": {"community": str(child_community.id)},
        "receiver": {"community": str(parent_community.id)},
        "created": res.json["created"],
        "updated": res.json["updated"],
        "expires_at": None,
        "type": "subcommunity",
        "status": "accepted",
        "title": 'Inclusion of subcommunity "Child Community"',
        "revision_id": 3,
        "is_closed": True,
        "is_expired": False,
        "is_open": False,
        "links": {
            "actions": {},
            "comments": f"https://127.0.0.1:5000/api/requests/{request_id}/comments",
            "self": f"https://127.0.0.1:5000/api/requests/{request_id}",
            "self_html": f"https://127.0.0.1:5000/requests/{request_id}",
            "timeline": f"https://127.0.0.1:5000/api/requests/{request_id}/timeline",
        },
    }

    # Check that the child community is a subcommunity of the parent community
    res = client.get(
        f"/communities/{str(child_community.id)}",
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 200
    assert res.json["parent"]["id"] == str(parent_community.id)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Subcommunities service schemas."""

from marshmallow import Schema, ValidationError, fields, post_load, pre_load


class MinimalCommunitySchema(Schema):
    """Minimal community schema."""

    slug = fields.String(required=True)
    title = fields.String(required=True)

    @post_load
    def load_default(self, data, **kwargs):
        """Load default values for a community."""
        return {
            "slug": data["slug"],
            "metadata": {
                "title": data["title"],
            },
            "access": {
                "visibility": "public",
                "members_visibility": "public",
                "member_policy": "open",
                "record_submission_policy": "open",
            },
        }


class SubcommunityRequestSchema(Schema):
    """Schema for subcommunity requests."""

    # Community input
    community_id = fields.String()
    community = fields.Nested(MinimalCommunitySchema)

    # community_id or community should be required
    @pre_load
    def validate(self, data, **kwargs):
        """Validate that either community_id or community is provided."""
        if "community_id" not in data and "community" not in data:
            raise ValidationError(
                "Either community_id or community should be provided."
            )
        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Subcommunities service errors."""


class SubCommunityError(Exception):
    """Base error for subcommunities service."""


class ParentChildrenNotAllowed(SubCommunityError):
    """Parent community does not accept subcommunities."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Community relations manager.

The manager provides the API to add, remove and iterate over communities
associated with a record.
"""

import uuid

from invenio_db import db
from invenio_records.api import Record

from invenio_communities.communities.records.api import Community
from invenio_communities.errors import SetDefaultCommunityError


class CommunitiesRelationManager:
    """Manager for a record's community relations."""

    def __init__(self, m2m_model_cls, record_id, data):
        """Constructor."""
        self._m2m_model_cls = m2m_model_cls
        self._record_id = record_id
        self._default_id = None
        self._communities_ids = set()
        self._communities_cache = {}
        self.from_dict(data)

    #
    # Helpers
    #
    def _to_id(self, val):
        """Get the community id."""
        if isinstance(val, (str, uuid.UUID)):
            return str(val)
        elif isinstance(val, Record):
            return str(val.id)
        return None

    def _lookup_community(self, community_id):
        """Retrieve a community by id.

        Caches the community.
        """
        if community_id not in self._communities_cache:
            c = Community.get_record(community_id)
            self._communities_cache[str(c.id)] = c
        return self._communities_cache[community_id]

    #
    # API
    #
    def add(self, community_or_id, request=None, default=False):
        """Add a record to a community.

        If a record was already added to a community an IntegrityError will
        be raised.
        """
        community_id = self._to_id(community_or_id)

        # Create M2M object
        obj = self._m2m_model_cls(
            record_id=self._record_id,
            community_id=community_id,
            request_id=self._to_id(request),
        )
        db.session.add(obj)

        # Add to internal set
        self._communities_ids.add(community_id)

        # Set default
        if default:
            self._default_id = community_id

        # Cache community only if provided
        if isinstance(community_or_id, Community):
            self._communities_cache[community_id] = community_or_id

    def remove(self, community_or_id):
        """Remove a record from a community."""
        community_id = self._to_id(community_or_id)

        # Delete M2M row.
        res = self._m2m_model_cls.query.filter_by(
            community_id=community_id, record_id=self._record_id
        ).delete()
        if res != 1:
            raise ValueError("The record has not been added to the community.")

        # Remove from internal set
        self._communities_ids.remove(community_id)

        # Unset default if needed
        if self._default_id == community_id:
            self._default_id = None

    def clear(self):
        """Clear all communities from the record."""
        # Remove all associations
        res = self._m2m_model_cls.query.filter_by(record_id=self._record_id).delete()
        self._communities_ids = set()
        self._default_id = None
        self._communities_cache = {}

    def refresh(self):
        """Refresh from the database M2M table."""
        # Retrieve from M2M table
        ids = (
            db.session.query(self._m2m_model_cls.community_id)
            .filter(self._m2m_model_cls.record_id == self._record_id)
            .all()
        )

        # Set internal list
        self._communities_ids = set([str(x[0]) for x in ids])

        # Unset default if no longer available
        if self._default_id and self._default_id not in self._communities_ids:
            self._default_id = None

    def __len__(self):
        """Get number of communities."""
        return len(self._communities_ids)

    def __contains__(self, community_or_id):
        """Check record is in community."""
        id_ = self._to_id(community_or_id)
        return id_ in self._communities_ids

    def __iter__(self):
        """Iterate over a communities."""
        # Determine community ids not already cached.
        nocache_ids = self._communities_ids - set(self._communities_cache.keys())

        # Fetch and cache missing community records
        if nocache_ids:
            communities = Community.get_records(nocache_ids)
            for c in communities:
                self._communities_cache[str(c.id)] = c

        # Iterate (sort by identifier to ensure consistent results)
        return (self._communities_cache[c] for c in sorted(self._communities_ids))

    @property
    def ids(self):
        """Get communities ids."""
        return sorted(self._communities_ids)

    @property
    def default(self):
        """Get the default community."""
        if self._default_id is not None:
            return self._lookup_community(self._default_id)
        return None

    @property
    def entries(self):
        """Get community objects list."""
        return list(self)

    @default.setter
    def default(self, community_or_id):
        """Set the default community.

        Note, the community must already have been added to the community. If
        not, then use ``.add(community, default=True)`` instead.
        """
        id_ = self._to_id(community_or_id)
        if id_ and id_ not in self._communities_ids:
            raise SetDefaultCommunityError
        self._default_id = id_

    @default.deleter
    def default(self):
        self._default_id = None

    # Persist relationships in record (denormalize the M2M table).
    # This enables 1) tracking community membership via the record versioning
    # and 2) faster indexing by not having to query the database for
    # relationships.
    def to_dict(self):
        """Get the dictionary which will be stored in the record."""
        data = {}
        if self._default_id is not None:
            data["default"] = self._default_id
        ids = list(self.ids)
        if ids:
            data["ids"] = ids
        return data

    def from_dict(self, data):
        """Build manager from the record dict."""
        data = data or {}
        self._default_id = data.get("default", None)
        self._communities_ids = set(data.get("ids", []))
        # Search results will have denormalized communities, so we can populate the
        # cache from it.
        entries = data.pop("entries", None)
        if entries:
            self._communities_cache = {c["id"]: Community.loads(c) for c in entries}
        return self



#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Update role_id type (downgrade recipe)."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "37b21951084c"
down_revision = "a3f5a8635cbb"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    pass


def downgrade():
    """Downgrade database."""
    op.alter_column(
        "communities_members",
        "group_id",
        existing_type=sa.String(length=80),
        type_=sa.INTEGER(),
        existing_nullable=True,
        postgresql_using="group_id::integer",
    )
    op.create_foreign_key(
        op.f("fk_communities_members_group_id_accounts_role"),
        "communities_members",
        "accounts_role",
        ["group_id"],
        ["id"],
        ondelete="RESTRICT",
    )
    op.alter_column(
        "communities_archivedinvitations",
        "group_id",
        existing_type=sa.String(length=80),
        type_=sa.INTEGER(),
        existing_nullable=True,
        postgresql_using="group_id::integer",
    )
    op.create_foreign_key(
        op.f("fk_communities_archivedinvitations_group_id_accounts_role"),
        "communities_archivedinvitations",
        "accounts_role",
        ["group_id"],
        ["id"],
        ondelete="RESTRICT",
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Subcommunities module."""

from .resources import SubCommunityResource, SubCommunityResourceConfig
from .services import SubCommunityService, SubCommunityServiceConfig

__all__ = (
    "SubCommunityResource",
    "SubCommunityResourceConfig",
    "SubCommunityService",
    "SubCommunityServiceConfig",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2024 CERN.
# Copyright (C) 2023      TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Communities Resource API config."""

import marshmallow as ma
from flask_resources import (
    HTTPJSONException,
    JSONSerializer,
    ResponseHandler,
    create_error_handler,
)
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.resources import RecordResourceConfig
from invenio_records_resources.resources.records.headers import etag_headers
from invenio_records_resources.services.base.config import ConfiguratorMixin, FromConfig
from invenio_requests.resources.requests.config import RequestSearchRequestArgsSchema

from invenio_communities.communities.resources.args import (
    CommunitiesSearchRequestArgsSchema,
)
from invenio_communities.communities.resources.serializer import (
    UICommunityJSONSerializer,
)
from invenio_communities.errors import (
    CommunityDeletedError,
    CommunityFeaturedEntryDoesNotExistError,
    LogoSizeLimitError,
    OpenRequestsForCommunityDeletionError,
    SetDefaultCommunityError,
)

community_error_handlers = RecordResourceConfig.error_handlers.copy()
community_error_handlers.update(
    {
        FileNotFoundError: create_error_handler(
            HTTPJSONException(
                code=404,
                description="No logo exists for this community.",
            )
        ),
        CommunityFeaturedEntryDoesNotExistError: create_error_handler(
            lambda e: HTTPJSONException(
                code=404,
                description=str(e),
            )
        ),
        LogoSizeLimitError: create_error_handler(
            lambda e: HTTPJSONException(
                code=400,
                description=str(e),
            )
        ),
        OpenRequestsForCommunityDeletionError: create_error_handler(
            lambda e: HTTPJSONException(
                code=400,
                description=str(e),
            )
        ),
        CommunityDeletedError: create_error_handler(
            lambda e: (
                HTTPJSONException(
                    code=410, description=_("The record has been deleted.")
                )
            )
        ),
        SetDefaultCommunityError: create_error_handler(
            lambda e: HTTPJSONException(
                code=400,
                description=str(e),
            )
        ),
    }
)


class CommunityResourceConfig(RecordResourceConfig, ConfiguratorMixin):
    """Communities resource configuration."""

    blueprint_name = "communities"
    url_prefix = ""

    routes = {
        "list": "/communities",
        "item": "/communities/<pid_value>",
        "rename": "/communities/<pid_value>/rename",
        "logo": "/communities/<pid_value>/logo",
        "featured-search": "/communities/featured",
        "featured-list": "/communities/<pid_value>/featured",
        "featured-item": "/communities/<pid_value>/featured/<featured_id>",
        "user-communities": "/user/communities",
        "community-requests": "/communities/<pid_value>/requests",
        "restore-community": "/communities/<pid_value>/restore",
        "list-subcommunities": "/communities/<pid_value>/subcommunities",
    }

    request_search_args = CommunitiesSearchRequestArgsSchema

    request_view_args = {
        **RecordResourceConfig.request_view_args,
        "featured_id": ma.fields.Int(),
    }
    error_handlers = FromConfig(
        "COMMUNITIES_ERROR_HANDLERS", default=community_error_handlers
    )
    request_community_requests_search_args = RequestSearchRequestArgsSchema

    response_handlers = {
        "application/json": ResponseHandler(JSONSerializer(), headers=etag_headers),
        "application/vnd.inveniordm.v1+json": ResponseHandler(
            UICommunityJSONSerializer(), headers=etag_headers
        ),
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Communities is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Notification related tests."""

from invenio_notifications.models import Notification

from invenio_communities.notifications.generators import CommunityMembersRecipient


def test_user_recipient_generator(community, members):
    """Add community members based on provided roles."""
    n = Notification(
        type="",
        context={"community": community.to_dict()},
    )

    # get all members of the community
    recipients = {}
    CommunityMembersRecipient(key="community")(n, recipients=recipients)
    assert len(recipients) == len(members)
    assert [u.id for u in members.values()] == list(recipients.keys())

    # get owner and reader only
    recipients = {}
    CommunityMembersRecipient(key="community", roles=["owner", "reader"])(
        n, recipients=recipients
    )
    assert len(recipients) == 2
    assert [members["owner"].id, members["reader"].id] == list(recipients)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2024 CERN.
# Copyright (C) 2023 Graz University of Technology.
# Copyright (C) 2023 KTH Royal Institute of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Default configuration."""

from datetime import timedelta

from invenio_i18n import lazy_gettext as _

from invenio_communities.communities.records.systemfields.access import (
    RecordSubmissionPolicyEnum,
)
from invenio_communities.communities.services import facets

COMMUNITIES_ROUTES = {
    "frontpage": "/communities",
    "search": "/communities/search",
    "new": "/communities/new",
    "upload": "/communities/<pid_value>/upload",
    "settings": "/communities/<pid_value>/settings",
    "requests": "/communities/<pid_value>/requests",
    "subcommunities": "/communities/<pid_value>/browse/subcommunities",
    "new_subcommunity": "/communities/<pid_value>/subcommunities/new",
    "settings_privileges": "/communities/<pid_value>/settings/privileges",
    "settings_submission_policy": "/communities/<pid_value>/settings/submission-policy",
    "settings_pages": "/communities/<pid_value>/settings/pages",
    "members": "/communities/<pid_value>/members",
    "invitations": "/communities/<pid_value>/invitations",
    "about": "/communities/<pid_value>/about",
    "curation_policy": "/communities/<pid_value>/curation-policy",
}

"""Communities ui endpoints."""

COMMUNITIES_FACETS = {
    "type": {
        "facet": facets.type,
        "ui": {
            "field": "type",
        },
    },
    "visibility": {
        "facet": facets.visibility,
        "ui": {
            "field": "visibility",
        },
    },
}
"""Available facets defined for this module."""


COMMUNITIES_SUBCOMMUNITIES_FACETS = COMMUNITIES_FACETS


COMMUNITIES_SORT_OPTIONS = {
    "bestmatch": dict(
        title=_("Best match"),
        fields=["_score"],  # ES defaults to desc on `_score` field
    ),
    "newest": dict(
        title=_("Newest"),
        fields=["-created"],
    ),
    "oldest": dict(
        title=_("Oldest"),
        fields=["created"],
    ),
    "version": dict(
        title=_("Version"),
        fields=["-versions.index"],
    ),
    "updated-desc": dict(
        title=_("Recently updated"),
        fields=["-updated"],
    ),
    "updated-asc": dict(
        title=_("Least recently updated"),
        fields=["updated"],
    ),
}
"""Definitions of available record sort options."""


COMMUNITIES_ROLES = [
    dict(
        name="reader",
        title=_("Reader"),
        description=_("Can view restricted records."),
        can_view=True,
    ),
    dict(
        name="curator",
        title=_("Curator"),
        description=_("Can curate records and view restricted records."),
        can_curate=True,
        can_view=True,
    ),
    dict(
        name="manager",
        title=_("Manager"),
        description=_(
            "Can manage members, curate records " "and view restricted records."
        ),
        can_manage_roles=["manager", "curator", "reader"],
        can_manage=True,
        can_curate=True,
        can_view=True,
    ),
    dict(
        name="owner",
        title=_("Owner"),
        description=_("Full administrative access to the entire community."),
        can_manage_roles=["owner", "manager", "curator", "reader"],
        is_owner=True,
        can_manage=True,
        can_curate=True,
        can_view=True,
    ),
]
"""Community roles."""

COMMUNITIES_SEARCH = {
    "facets": ["type", "visibility"],
    "sort": ["bestmatch", "newest", "oldest"],
}
"""Community search configuration (i.e list of communities)"""

COMMUNITIES_SEARCH_SORT_BY_VERIFIED = False
"""Sort communities by 'verified' first."""

COMMUNITIES_SUBCOMMUNITIES_SEARCH = {
    "facets": ["type"],
    "sort": ["bestmatch", "newest", "oldest"],
}
"""Community search configuration (i.e list of communities)"""

COMMUNITIES_REQUESTS_SEARCH = {
    "facets": ["type", "status"],
    "sort": ["bestmatch", "newest", "oldest"],
}
"""Community requests search configuration (i.e list of community requests)"""

COMMUNITIES_MEMBERS_SEARCH = {
    "facets": ["role", "visibility"],
    "sort": ["bestmatch", "name", "newest", "oldest"],
}
"""Community requests search configuration (i.e list of community requests)"""

COMMUNITIES_MEMBERS_SORT_OPTIONS = {
    "bestmatch": dict(
        title=_("Best match"),
        fields=["_score"],  # ES defaults to desc on `_score` field
    ),
    "name": dict(
        title=_("Name"),
        fields=["user.profile.full_name.keyword"],
    ),
    "newest": dict(
        title=_("Newest"),
        fields=["-created"],
    ),
    "oldest": dict(
        title=_("Oldest"),
        fields=["created"],
    ),
}
"""Definitions of available record sort options."""

COMMUNITIES_MEMBERS_FACETS = {
    "role": {
        "facet": facets.role,
        "ui": {
            "field": "role",
        },
    },
    "visibility": {
        "facet": facets.visible,
        "ui": {
            "field": "visible",
        },
    },
}
"""Available facets defined for this module."""

COMMUNITIES_INVITATIONS_SEARCH = {
    "facets": ["type", "status"],
    "sort": ["bestmatch", "name", "newest", "oldest"],
}
"""Community invitations search configuration (i.e list of community invitations)"""

COMMUNITIES_INVITATIONS_SORT_OPTIONS = {
    "bestmatch": dict(
        title=_("Best match"),
        fields=["_score"],  # ES defaults to desc on `_score` field
    ),
    "name": dict(
        title=_("Name"),
        fields=["user.profile.full_name.keyword"],
    ),
    "newest": dict(
        title=_("Newest"),
        fields=["-created"],
    ),
    "oldest": dict(
        title=_("Oldest"),
        fields=["created"],
    ),
}
"""Definitions of available record sort options."""


COMMUNITIES_INVITATIONS_EXPIRES_IN = timedelta(days=30)
""""Default amount of time before an invitation expires."""

COMMUNITIES_LOGO_MAX_FILE_SIZE = 10**6
"""Community logo size quota, in bytes."""


COMMUNITIES_NAMESPACES = {}
"""Custom fields namespaces.

.. code-block:: python
    {<namespace>: <uri>, ...}

For example:

.. code-block:: python
    {
        "cern": "https://cern.ch/terms",
        "dwc": "http://rs.tdwg.org/dwc/terms/"
    }

"""

COMMUNITIES_CUSTOM_FIELDS = []
"""Communities custom fields definition.

Of the shape:

.. code-block:: python

    [
        <custom-field-class-type(name='field')>,
        # ...
        <custom-field-class-type(name='fieldN')>'
    ]

For example:

.. code-block:: python

    [
        TextCF(name="experiment"),
        ...
    ]
"""

COMMUNITIES_CUSTOM_FIELDS_UI = []
"""Communities custom fields UI configuration.

Of the shape:

.. code-block:: python

    [{
        section: <section_name>,
        fields: [
            {
                field: "path-to-field",  # this should be validated against the defined fields in `RDM_CUSTOM_FIELDS`
                ui_widget: "<ui-widget-name>",  # predefined or user defined ui widget
                props: {
                    label:"<ui-label-to-display>",
                    placeholder:"<placeholder-passed-to-widget>",
                    icon:"<icon-passed-to-widget>",
                    description:"<description-passed-to-widget>",
                }
            },
        ],

        # ...
    }]

For example:

.. code-block:: python

    [{
        "section": "CERN Experiment"
        "fields" : [{
            field: "experiment",  # this should be validated against the defined fields in `RDM_CUSTOM_FIELDS`
            ui_widget: "CustomTextField",  # user defined widget in my-site
            props: {
                label: "Experiment",
                placeholder: "Type an experiment...",
                icon: "pencil",
                description: "You should fill this field with one of the experiments e.g LHC, ATLAS etc.",
            }
        },
        ...
    }]
"""

# Preview: Feature flag for invenio-app-rdm v11
COMMUNITIES_ADMINISTRATION_DISABLED = True

COMMUNITIES_ALLOW_RESTRICTED = True

# Cache duration
# 60 seconds * 60 (1 hour) * 24 (1 day)
COMMUNITIES_IDENTITIES_CACHE_TIME = 60 * 60 * 24

# Redis URL Cache for identities
COMMUNITIES_IDENTITIES_CACHE_REDIS_URL = "redis://localhost:6379/4"

# Cache handler
COMMUNITIES_IDENTITIES_CACHE_HANDLER = (
    "invenio_communities.cache.redis:IdentityRedisCache"
)

COMMUNITIES_OAI_SETS_PREFIX = "community-"

COMMUNITIES_ALWAYS_SHOW_CREATE_LINK = False
"""Controls visibility of 'New Community' btn based on user's permission when set to True."""

COMMUNITIES_ALLOW_MEMBERSHIP_REQUESTS = False
"""Feature flag for membership request."""

COMMUNITIES_DEFAULT_RECORD_SUBMISSION_POLICY = RecordSubmissionPolicyEnum.OPEN
"""Default value of record submission policy community access setting."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# Invenio-RDM-Records is free software; you can redistribute it
# and/or modify it under the terms of the MIT License; see LICENSE file for
# more details.

"""Test permissions."""

from invenio_communities.permissions import CommunityPermissionPolicy


def test_can_request_membership(
    app, community, owner, anon_identity, any_user, superuser_identity
):

    policy = CommunityPermissionPolicy
    community_record = community._record
    authenticated_identity = any_user.identity

    allow_membership_requests_orig = app.config["COMMUNITIES_ALLOW_MEMBERSHIP_REQUESTS"]

    # Case - feature disabled
    app.config["COMMUNITIES_ALLOW_MEMBERSHIP_REQUESTS"] = False
    assert not (
        policy(action="request_membership", record=community_record).allows(
            superuser_identity
        )
    )

    app.config["COMMUNITIES_ALLOW_MEMBERSHIP_REQUESTS"] = True

    # Case - setting disabled
    community_record.access.member_policy = "closed"
    assert not (
        policy(action="request_membership", record=community_record).allows(
            superuser_identity
        )
    )

    community_record.access.member_policy = "open"

    # Case - unlogged user
    assert not (
        policy(action="request_membership", record=community_record).allows(
            anon_identity
        )
    )

    # Case - logged user not part of community
    assert policy(action="request_membership", record=community_record).allows(
        authenticated_identity
    )

    # Case - member of community
    assert not (
        policy(action="request_membership", record=community_record).allows(
            owner.identity
        )
    )

    app.config["COMMUNITIES_ALLOW_MEMBERSHIP_REQUESTS"] = allow_membership_requests_orig
